/*      config.c  -  Display a user interface for setting configuration options */

#include "config.h"
#define DYNAMIC_ROMS 1
#include "memu.h"
#include "monprom.h"
#include "win.h"
#include "kbd.h"
#include "mem.h"
#include "vid.h"
#include "mon.h"
#include "snd.h"
#include "sdxfdc.h"
#include "common.h"
#include <dirt.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#define CLR_BACKGROUND     0
#define CLR_HIGHBACK       1
#define CLR_NORMAL         2
#define CLR_HIGHLIGHT      3
#define CLR_DISABLED       4
#define CLR_HELP           5
#define CLR_COUNT          6

#define STY_NORMAL         ( 16 * CLR_NORMAL    + CLR_BACKGROUND )
#define STY_HIGHLIGHT      ( 16 * CLR_HIGHLIGHT + CLR_HIGHBACK )
#define STY_DISABLED       ( 16 * CLR_DISABLED  + CLR_BACKGROUND )
#define STY_HELP           ( 16 * CLR_HELP      + CLR_BACKGROUND )

#define WINCFG_WTH         80
#define WINCFG_HGT         24

#define CFG_MTX500         0
#define CFG_MTX512         1
#define CFG_SDX            2
#define CFG_CPM_MONO       3
#define CFG_CPM_COLOUR     4
#define CFG_COUNT          5

#define SDX_MEM_BLOCKS     ( ( 512 + 64 ) / 16 )

#define EXIT_APPLY         0
#define EXIT_CANCEL        1
#define EXIT_RESTART       2
#define EXIT_EXIT          3
#define EXIT_COUNT         4

#define STATE_FOCUS        2
#define STATE_NORMAL       1
#define STATE_DISABLED     0

#define ROW_CFG            0
#define ROW_KEYMAP         1
#define ROW_TAPE           2
#define ROW_DRIVE_B        3
#define ROW_DRIVE_C        4
#define ROW_EXIT           5
#define ROW_HELP           6
#define ROW_CHOOSE         7

/* The following defines initial allocation for directory listing.
   This is then doubled as often as necessary, c.f STDC++ Vector container */

#define INIT_NFILE         20
#define MAX_PATH           260

static WIN *cfg_win  =  NULL;
static int  cfg_wk   =  -1;
static COL cfg_clr[CLR_COUNT] = {
	{   0,   0,   0 },   /* Backgound */
	{  64,  64,  64 },   /* Highlighted background */
	{   0, 255, 255 },   /* Normal text */
	{ 255, 255,   0 },   /* Highlighted text */
	{   0, 127, 127 },   /* Disabled text */
	{   0, 255,   0 }    /* Help text */
	};

static char *  config_fn      =  NULL;
static char *  disk_dir       =  NULL;
static int     rom_enable     =  0xff;
static BOOLEAN bCfgRedraw     =  TRUE;
static int     iCfgOld        =  0;
static int     iCfgCur        =  0;
static int     iCfgSel        =  0;
static BOOLEAN bCfgRemap      =  FALSE;
static BOOLEAN bSelRemap      =  FALSE;
static char *  psCfgTape      =  NULL;
static char *  psCurTape      =  NULL;
static char *  psCfgDirB      =  NULL;
static char *  psCfgDriveB    =  NULL;
static char *  psCfgDirC      =  NULL;
static char *  psCfgDriveC    =  NULL;
static char *  psCurDrive[2]  =  { NULL, NULL };
static int     iCfgExit       =  0;
static BOOLEAN bCfgExit       =  FALSE;

void config_term (void)
	{
	if ( cfg_win != NULL )
		{
		win_delete (cfg_win);
		cfg_win  =  NULL;
		}
	}

static void cfg_keypress(int wk)
	{
	cfg_wk =  wk;
	}

static void cfg_keyrelease(int wk)
	{
	}

static void cfg_print (int iRow, int iCol, int iSty, const char *psTxt, int nCh)
	{
	BOOLEAN  bEOT;
	int  iFG =  iSty >> 4;
	int  iBG =  iSty & 0x0f;
	byte *pby   =  cfg_win->data + GLYPH_HEIGHT * cfg_win->width * iRow + GLYPH_WIDTH * iCol;
	char ch;
	byte by;
	int  iScan, iPix, iCh;
	if ( nCh <= 0 )   nCh   =  strlen (psTxt);

	for ( iScan = 0; iScan < GLYPH_HEIGHT; ++iScan )
		{
		bEOT  =  FALSE;
		for ( iCh = 0; iCh < nCh; ++iCh )
			{
			if ( ! bEOT )
				{
				ch =  psTxt[iCh];
				if ( ch == '\0' )
					{
					bEOT  =  TRUE;
					ch =  ' ';
					}
				else if ( ch < 0 )   ch =  '?';
				}
			else  ch =  ' ';
			by =  mon_alpha_prom[ch][iScan];
			for ( iPix = 0; iPix < GLYPH_WIDTH; ++iPix )
				{
				*pby  =  ( by & 0x80 ) ? iFG : iBG;
				++pby;
				by = by << 1;
				}
			}
		pby   += cfg_win->width - GLYPH_WIDTH * nCh;
		}
	}

static void cfg_help (const char *psHelp)
	{
	cfg_print (ROW_HELP, 0, STY_HELP, psHelp, WINCFG_WTH);
	}

static void cfg_csr (int iRow, int iCol, int iSty)
	{
	int  iFG =  iSty >> 4;
	int  iBG =  iSty & 0x0f;
	int  iPix;
	byte *pby   =  cfg_win->data + ( GLYPH_HEIGHT * ( iRow + 1 ) - 1 ) * cfg_win->width + GLYPH_WIDTH * iCol;
	for ( iPix = 0; iPix < GLYPH_WIDTH; ++iPix )
		{
		if ( *pby == iFG )   *pby  =  iBG;
		else                 *pby  =  iFG;
		++pby;
		}
	}

static void cfg_clear_rows (int iFirst, int iLast)
	{
	memset (cfg_win->data + GLYPH_HEIGHT * cfg_win->width * iFirst,
		CLR_BACKGROUND,
		GLYPH_HEIGHT * cfg_win->width * ( iLast - iFirst ));
	}

static int cfg_key (void)
	{
	win_refresh (cfg_win);
	cfg_wk   =  -1;
	while ( cfg_wk < 0 )
		win_handle_events ();
	return   cfg_wk;
	}

static int cfg_edit (int iRow, int iCol, int nWth, int iSty, int nLen, char *psText)
	{
	int  nCh    =  strlen (psText);
	int  iCsr   =  0;
	int  iScl   =  0;
	int  wk;
	while (TRUE)
		{
		if ( ( iCsr - iScl ) >= nWth )   iScl  =  iCsr - nWth + 1;
		else if ( iCsr < iScl )          iScl  =  iCsr;
		cfg_print (iRow, iCol, iSty, &psText[iScl], nWth);
		cfg_csr (iRow, iCol + iCsr - iScl, iSty);
		wk =  cfg_key ();
		switch (wk)
			{
			case  WK_Return:
			case  WK_Escape:
			{
			return   wk;
			}
			case  WK_Left:
			{
			if ( iCsr > 0 )   --iCsr;
			break;
			}
			case  WK_Home:
			{
			iCsr  =  0;
			break;
			}
			case  WK_End:
			{
			iCsr  =  nCh;
			break;
			}
			case  WK_Right:
			{
			if ( iCsr < nCh ) ++iCsr;
			break;
			}
			case  WK_Delete:
			{
			if ( iCsr < nCh )
				{
				strcpy (&psText[iCsr], &psText[iCsr+1]);
				--nCh;
				}
			break;
			}
			case  WK_BackSpace:
			{
			if ( iCsr > 0 )
				{
				--iCsr;
				strcpy (&psText[iCsr], &psText[iCsr+1]);
				--nCh;
				}
			break;
			}
			default:
			{
			if ( ( nCh < nLen - 1 ) && ( wk >= ' ' ) && ( wk <= '~' ) )
				{
				memmove (&psText[iCsr+1], &psText[iCsr], nCh + 1 - iCsr);
				psText[iCsr]   =  wk;
				++nCh;
				++iCsr;
				}
			break;
			}
			}
		}
	}

static void row_cfg_draw (int iState)
	{
/*               1         2         3         4         5         6         7
				 01234567890123456789012345678901234567890123456789012345678901234567890123456789 */
	static char *psHelp[CFG_COUNT]   =
		{
		"<Space> to select: Memotech MTX500 with 32k byte of RAM",
		"<Space> to select: Memotech MTX512 with 64k byte of RAM",
		"<Space> to select: Memotech with 576k bytes of RAM and two type 07 disk drives",
		"<Space> to select: 576k bytes of RAM, two type 07 drives, CP/M, mono monitor",
		"<Space> to select: 576k bytes of RAM, two type 07 drives, CP/M, colour monitor"
		};
	static char sCfg[]   =  " [ ] MTX500     [ ] MTX512     [ ] SDX        [ ] CPM MONO   [ ] CPM COLOUR ";
	sCfg[ 2] =  ( iCfgCur == CFG_MTX500     ) ? '*' : ' ';
	sCfg[17] =  ( iCfgCur == CFG_MTX512     ) ? '*' : ' ';
	sCfg[32] =  ( iCfgCur == CFG_SDX        ) ? '*' : ' ';
	sCfg[47] =  ( iCfgCur == CFG_CPM_MONO   ) ? '*' : ' ';
	sCfg[62] =  ( iCfgCur == CFG_CPM_COLOUR ) ? '*' : ' ';
	cfg_print (ROW_CFG, 0, STY_NORMAL, sCfg, 0);
	if ( iState == STATE_FOCUS )
		{
		cfg_print (ROW_CFG, 15*iCfgSel, STY_HIGHLIGHT, &sCfg[15*iCfgSel], 15);
		cfg_help (psHelp[iCfgSel]);
		}
	}

static BOOLEAN row_cfg_key (int wk)
	{
	switch (wk)
		{
		case  WK_Left:
		{
		if ( iCfgSel == 0 )  iCfgSel  =  CFG_COUNT;
		--iCfgSel;
		break;
		}
		case  WK_Right:
		{
		++iCfgSel;
		if ( iCfgSel == CFG_COUNT )  iCfgSel  =  0;
		break;
		}
		case  WK_Return:
		case  ' ':
		{
		iCfgCur  =  iCfgSel;
		bCfgRedraw  =  TRUE;
		break;
		}
		default:
		{
		return   FALSE;
		}
		}
	return   TRUE;
	}

static void row_kbd_draw (int iState)
	{
	static char sNomap[] =  " [ ] Normal keyboard  ";
	static char sRemap[] =  " [ ] Rempped keyboard ";
	if ( bCfgRemap )
		{
		sNomap[2]   =  ' ';
		sRemap[2]   =  '*';
		}
	else
		{
		sNomap[2]   =  '*';
		sRemap[2]   =  ' ';
		}
	if ( iState != STATE_FOCUS )
		{
		cfg_print (ROW_KEYMAP, 0, STY_NORMAL, sNomap, 0);
		cfg_print (ROW_KEYMAP, 22, STY_NORMAL, sRemap, 0);
		}
	else if ( bSelRemap )
		{
		cfg_print (ROW_KEYMAP, 0, STY_NORMAL, sNomap, 0);
		cfg_print (ROW_KEYMAP, 22, STY_HIGHLIGHT, sRemap, 0);
		cfg_help ("<Space> to select: Keys match PC keyboard - Problems with some games");
		}
	else
		{
		cfg_print (ROW_KEYMAP, 0, STY_HIGHLIGHT, sNomap, 0);
		cfg_print (ROW_KEYMAP, 22, STY_NORMAL, sRemap, 0);
		cfg_help ("<Space> to select: Keys match MTX keyboard - Not always PC key symbols");
		}
	}

static BOOLEAN row_kbd_key (int wk)
	{
	switch (wk)
		{
		case  WK_Left:
		case  WK_Right:
		{
		bSelRemap   =  !  bSelRemap;
		break;
		}
		case  WK_Return:
		case  ' ':
		{
		bCfgRemap   =  bSelRemap;
		break;
		}
		default:
		{
		return   FALSE;
		}
		}
	return   TRUE;
	}

static int cfg_choose (int nItem, char **ppsItem)
	{
	int  nLenMax   =  0;
	int  nLen, i;
	int  nCol, nWth, nRow;
	int  iChoose, iScroll, iItem;
	int  iRow, iCol;
	int  nVis   =  WINCFG_HGT - ROW_CHOOSE;
	cfg_help ("Cursor keys to move highlight, <Space> to select, <Esc> to cancel");
	for ( i = 0; i < nItem; ++i )
		{
		nLen  =  strlen (ppsItem[i]);
		if ( nLen > nLenMax )   nLenMax  =  nLen;
		}
	if ( nLenMax + 2 < WINCFG_WTH )  nCol  =  WINCFG_WTH / ( nLenMax + 2 );
	else                             nCol  =  1;
	nWth =   WINCFG_WTH / nCol;
	nRow =   ( nItem - 1 ) / nCol + 1;
	iChoose  =  0;
	iScroll  =  0;
	while (TRUE)
		{
		iItem    =  iScroll;
		for ( iRow = 0; ( iRow < nVis ) && ( iItem < nItem ); ++iRow )
			{
            for ( iCol = 0; ( iCol < nCol ) && ( iItem < nItem ); ++iCol, ++iItem )
				{
				if ( iItem == iChoose )
					cfg_print (iRow + ROW_CHOOSE, iCol * nWth, STY_HIGHLIGHT, ppsItem[iItem], nWth);
				else
					cfg_print (iRow + ROW_CHOOSE, iCol * nWth, STY_NORMAL, ppsItem[iItem], nWth);
				}
			}
		switch (cfg_key ())
			{
			case  WK_Right:
			{
			if ( iChoose < nItem - 1 )
				{
				++iChoose;
				if ( ( iChoose - iScroll ) / nCol >= nVis )  iScroll  += nCol;
				}
			break;
			}
			case  WK_Left:
			{
			if ( iChoose > 0 )
				{
				--iChoose;
				if ( iChoose < iScroll )   iScroll  -= nCol;
				}
			break;
			}
			case  WK_Down:
			{
			if ( iChoose < nItem - nCol )
				{
				iChoose  += nCol;
				if ( ( iChoose - iScroll ) / nCol >= nVis )  iScroll  += nCol;
				}
			break;
			}
			case  WK_Up:
			{
			if ( iChoose >= nCol )
				{
				iChoose  -= nCol;
				if ( iChoose < iScroll )   iScroll  -= nCol;
				}
			break;
			}
			case  WK_Return:
			case  ' ':
			{
			cfg_clear_rows (ROW_CHOOSE, WINCFG_HGT);
			return   iChoose;
			}
			case  WK_Escape:
			{
			cfg_clear_rows (ROW_CHOOSE, WINCFG_HGT);
			return   -1;
			}
			}
		}
	}

static char *cfg_new_file (const char *psDir, const char *psExt)
	{
	char *psPath;
	char sFile[MAX_PATH];
	strncpy (sFile, psExt, MAX_PATH);
	sFile[MAX_PATH-1]   =  '\0';
	cfg_help ("Enter new file name below, <Return> to finish, <Esc> to cancel");
	cfg_print (ROW_CHOOSE, 0, STY_NORMAL, "New file:", 0);
	if ( cfg_edit (ROW_CHOOSE, 10, WINCFG_WTH - 10, STY_NORMAL, MAX_PATH, sFile) == WK_Return )
		{
		FILE  *pfil;
		int   nDir  =  strlen (psDir);
		int   nFile =  strlen (sFile);
		int   nExt  =  strlen (psExt);
		if ( ( nFile < nExt ) || ( strcmp (&sFile[nFile - nExt], psExt) != 0 ) )
			{
			if ( nFile + nExt > MAX_PATH - 1 )  strcpy (&sFile[MAX_PATH - 1 - nExt], psExt);
			else                                strcpy (&sFile[nFile], psExt);
			nFile =  strlen (sFile);
			}
		psPath   =  (char *) emalloc (nDir + nFile + 2);
		strcpy (psPath, psDir);
		psPath[nDir]   =  '/';
		strcpy (&psPath[nDir+1], sFile);
		pfil  =  fopen (psPath, "wb");
		if ( pfil == NULL )  fatal ("Failed to create file %s", psPath);
		fclose (pfil);
		return   estrdup (sFile);
		}
	return   NULL;
	}

static char *cfg_choose_file (const char *psDir, const char *psExt, BOOLEAN bDisk)
	{
	char *psFile;
	const char *psFile2;
	char **ppsFiles   =  (char **) malloc (INIT_NFILE * sizeof (char *));
	int  nAlloc       =  INIT_NFILE;
	int  nFiles       =  0;
	int  nCh;
	int  nExt         =  strlen (psExt);
	int  i;
	int  iErr;
	int  iChoose;
	DIRT *dirt;
	dirt =  dirt_open (psDir, &iErr);
	if ( iErr != DIRTE_OK ) fatal (dirt_error (iErr));
	if ( bDisk )
		{
		nFiles   =  2;
		ppsFiles[0] =  estrdup ("<None>");
		ppsFiles[1] =  estrdup ("<New>");
		}
	while ( ( psFile2 = dirt_next (dirt) ) != NULL )
		{
		nCh   =  strlen (psFile2);
		if ( ( nCh > nExt ) && ( strcmp (&psFile2[nCh-nExt], psExt) == 0 ) )
			{
			if ( nFiles >= nAlloc )
				{
				nAlloc   *= 2;
				ppsFiles =  (char **) realloc (ppsFiles, nAlloc * sizeof (char *));
				if ( ppsFiles == NULL ) fatal ("Memory allocation failure");
				}
			ppsFiles[nFiles]  =  estrdup (psFile2);
			++nFiles;
			}
		}
	dirt_close (dirt);
	iChoose  =  cfg_choose (nFiles, ppsFiles);
	if ( iChoose >= 0 )  psFile   =  ppsFiles[iChoose];
	else                 psFile   =  NULL;
	for ( i = 0; i < nFiles; ++i )
		{
		if ( ppsFiles[i] != psFile )  free (ppsFiles[i]);
		}
	free (ppsFiles);
	if ( ( bDisk ) && ( iChoose == 1 ) )
		{
		free (psFile);
		psFile   =  cfg_new_file (psDir, psExt);
		}
	return   psFile;
	}

static void file_draw (int iRow, const char *psTitle, const char *psFile, int iState)
	{
	char sLine[WINCFG_WTH+1];
	int  nCh =  strlen (psTitle);
	strcpy (sLine, psTitle);
	if ( psFile != NULL )
		{
		strncpy (&sLine[nCh], psFile, WINCFG_WTH - nCh);
		sLine[WINCFG_WTH] =  '\0';
		}
	if ( iState == STATE_FOCUS )
		{
		cfg_print (iRow, 0, STY_HIGHLIGHT, sLine, WINCFG_WTH);
		cfg_help ("<Space> to change selected file");
		}
	else if ( iState == STATE_NORMAL )
		{
		cfg_print (iRow, 0, STY_NORMAL, sLine, WINCFG_WTH);
		}
	else
		{
		cfg_print (iRow, 0, STY_DISABLED, sLine, WINCFG_WTH);
		}
	}

static BOOLEAN file_key (const char *psDir, const char *psExt, BOOLEAN bDisk, char **ppsFile, int wk)
	{
	if ( ( wk == WK_Return ) || ( wk == ' ' ) )
		{
		char *psFile   =  cfg_choose_file (psDir, psExt, bDisk);
		if ( psFile != NULL )
			{
			if ( *ppsFile != NULL ) free (*ppsFile);
			*ppsFile =  psFile;
			}
		return   TRUE;
		}
	return   FALSE;
	}

static void row_tape_draw (int iState)
	{
	file_draw (ROW_TAPE, "Tape: ", psCfgTape, iState);
	}

static BOOLEAN row_tape_key (int wk)
	{
	return   file_key (cfg.tape_name_prefix, ".mtx", FALSE, &psCfgTape, wk);
	}

static void row_drive_b_draw (int iState)
	{
	file_draw (ROW_DRIVE_B, "Drive B: ", psCfgDriveB, iState);
	}

static BOOLEAN row_drive_b_key (int wk)
	{
	return   file_key (psCfgDirB, ".mfloppy", TRUE, &psCfgDriveB, wk);
	}

static void row_drive_c_draw (int iState)
	{
	file_draw (ROW_DRIVE_C, "Drive C: ", psCfgDriveC, iState);
	}

static BOOLEAN row_drive_c_key (int wk)
	{
	return   file_key (psCfgDirC, ".mfloppy", TRUE, &psCfgDriveC, wk);
	}

static void cfg_set_disk_dir (const char *psDir)
   {
   if ( disk_dir != NULL )  free (disk_dir);
   disk_dir =  NULL;
   if ( psDir != NULL ) disk_dir  =  estrdup (psDir);
   }

static void cfg_get_drive (int iDrive, char **ppsDir, char **ppsFile)
	{
	const char *ps1   =  cfg.fn_sdxfdc[iDrive];
	if ( *ppsFile != NULL ) free (*ppsFile);
	if ( ( ps1 == NULL ) || ( ps1[0] == '\0' ) )
		{
		*ppsFile =  estrdup ("<None>");
		}
	else
		{
		char *ps2   =  strrchr (ps1, '/');
		if ( ps2 != NULL )
			{
			int   nDir  =  ps2 - ps1;
			if ( *ppsDir != NULL ) free (*ppsDir);
			*ppsDir  =  (char *) emalloc (nDir + 1);
			strncpy (*ppsDir, ps1, nDir);
			(*ppsDir)[nDir]   =  '\0';
			ps1   =  ps2 + 1;
			if ( disk_dir == NULL )   cfg_set_disk_dir (*ppsDir);
			}
		else if ( *ppsDir == NULL )
			{
			const char *ps2   =  disk_dir;
			if ( ps2 != NULL )   *ppsDir  =  estrdup (ps2);
			}
		*ppsFile =  estrdup (ps1);
		}
	}

static void cfg_set_drive (int iDrive, const char *psDir, const char *psFile)
	{
	if ( ( psFile == NULL ) || ( psFile[0] == '\0' ) || ( strcmp (psFile, "<None>") == 0 ) )
		{
		sdxfdc_init (iDrive, NULL);
		if ( psCurDrive[iDrive] != NULL ) free (psCurDrive[iDrive]);
		psCurDrive[iDrive] = NULL;
		cfg.fn_sdxfdc[iDrive] = NULL;
		}
	else
		{
		const char *psOld =  cfg.fn_sdxfdc[iDrive];
		char *psPath   =  emalloc (strlen (psDir) + strlen (psFile) + 2);
		strcpy (psPath, psDir);
		strcat (psPath, "/");
		strcat (psPath, psFile);
		if ( ( psOld == NULL ) || ( strcmp (psPath, psOld) != 0 ) )
			{
			sdxfdc_init (iDrive, psPath);
			if ( psCurDrive[iDrive] != NULL ) free (psCurDrive[iDrive]);
			psCurDrive[iDrive] = psPath;
			cfg.fn_sdxfdc[iDrive] = psPath;
			}
		else
			{
			free (psPath);
			}
		}
	}

static void cfg_set_tape (const char *psTape)
	{
	if ( psCurTape != NULL ) free (psCurTape);
	psCurTape = NULL;
	if ( psTape != NULL ) psCurTape = estrdup (psTape);
	cfg.tape_fn = psCurTape;
	}

static void cfg_init (void)
	{
	const char *psDir;
	if ( ( rom_enable & ROMEN_CPM ) && ( cfg.rom_fn[ROM_CPM] != NULL ) )
		{
		if ( cfg.mon_emu & MONEMU_WIN_MONO ) iCfgOld  =  CFG_CPM_MONO;
		else                                 iCfgOld  =  CFG_CPM_COLOUR;
		}
	else if ( ( rom_enable & ROMEN_SDX2 ) && ( cfg.rom_fn[ROM_SDX2] != NULL ) )
		{
		iCfgOld  =  CFG_SDX;
		}
	else if ( mem_get_alloc () >= 4 )
		{
		iCfgOld  =  CFG_MTX512;
		}
	else
		{
		iCfgOld  =  CFG_MTX500;
		}
	iCfgCur  =  iCfgOld;
	iCfgSel  =  0;

	bCfgRemap      =  cfg.kbd_emu & KBDEMU_REMAP;
	bSelRemap      =  FALSE;
	if ( psCfgTape != NULL )   free (psCfgTape);
	psCfgTape = NULL;
	if ( cfg.tape_fn != NULL )   psCfgTape   =  estrdup (cfg.tape_fn);
	cfg_get_drive (0, &psCfgDirB, &psCfgDriveB);
	cfg_get_drive (1, &psCfgDirC, &psCfgDriveC);
	psDir =  disk_dir;
	if ( psDir != NULL )
		{
		if ( psCfgDirB == NULL )   psCfgDirB   =  estrdup (psDir);
		if ( psCfgDirC == NULL )   psCfgDirC   =  estrdup (psDir);
		}
	iCfgExit =  0; 

	cfg_win  =  win_create (GLYPH_WIDTH * WINCFG_WTH, GLYPH_HEIGHT * WINCFG_HGT,
		cfg.mon_width_scale, cfg.mon_height_scale,
		"MEMU Configuration",
		NULL, NULL,
		cfg_clr, CLR_COUNT,
		cfg_keypress, cfg_keyrelease);
	}

static void cfg_save (const char *psConfig)
	{
	if ( psConfig != NULL )
		{
		const char *ps;
		int   rom;
		FILE *pfil  =  fopen (psConfig, "w");
		if ( pfil == NULL )  fatal ("Unable to write configuration file");
		fprintf (pfil, "-mem-blocks %d\n", mem_get_alloc ());
		for ( rom = 0; rom < 8; ++rom )
			{
			if ( cfg.rom_fn[rom] != NULL ) fprintf (pfil, "-rom%d \"%s\"\n", rom, cfg.rom_fn[rom]);
			}
		if ( cfg.kbd_emu & KBDEMU_REMAP ) fprintf (pfil, "-kbd-remap\n");
		fprintf (pfil, "-kbd-country %d\n", ( cfg.kbd_emu & KBDEMU_COUNTRY ) >> 2);
		if ( cfg.vid_emu & VIDEMU_WIN )
			{
			int   scale = cfg.vid_height_scale;
			fprintf (pfil, "-vid-win\n");
			while ( --scale > 0 )
				{
				fprintf (pfil, "-vid-win-big\n");
				}
			if ( cfg.vid_emu & VIDEMU_WIN_HW_PALETTE ) fprintf (pfil, "-vid-win-hw-palette\n");
			}
		if ( cfg.snd_emu & SNDEMU_PORTAUDIO )
			{
			fprintf (pfil, "-snd-portaudio\n");
			if ( cfg.latency != 0.0 )   fprintf (pfil, "-snd-latency %f\n", cfg.latency);
			}
		if ( cfg.mon_emu & MONEMU_WIN )
			{
			int   scale = cfg.mon_height_scale;
			fprintf (pfil, "-mon-win\n");
			while ( --scale > 0 )
				{
				fprintf (pfil, "-mon-win-big\n");
				}
			if ( cfg.mon_emu & MONEMU_WIN_MONO ) fprintf (pfil, "-mon-win-mono\n");
			if ( cfg.mon_emu & MONEMU_IGNORE_INIT ) fprintf (pfil, "-mon-ignore-init\n");
			}
		else
			{
			if ( cfg.mon_height_scale > 1 )  fprintf (pfil, "-mon-size %d\n", cfg.mon_height_scale);
			}
		if ( disk_dir != NULL ) fprintf (pfil, "-disk-dir \"%s\"\n", disk_dir);
		if ( cfg.fn_sdxfdc[0] != NULL ) fprintf (pfil, "-sdx-mfloppy \"%s\"\n", cfg.fn_sdxfdc[0]);
		if ( cfg.fn_sdxfdc[1] != NULL ) fprintf (pfil, "-sdx-mfloppy2 \"%s\"\n", cfg.fn_sdxfdc[1]);
		if ( cfg.tape_name_prefix != NULL ) fprintf (pfil, "-tape-dir \"%s\"\n", cfg.tape_name_prefix);
		rom_enable   =  mem_get_rom_enable ();
		if ( rom_enable != 0xff )   fprintf (pfil, "-rom-enable 0x%02x\n", rom_enable);
		fclose (pfil);
		}
	}

static void cfg_restart (void)
	{
	if ( iCfgCur != iCfgOld )
		{
		mon_term ();
		switch ( iCfgCur )
			{
			case  CFG_MTX500:
			{
			if ( cfg.mon_emu & ( MONEMU_WIN | MONEMU_TH | MONEMU_CONSOLE ) )
				{
				mon_term ();
				cfg.mon_emu &= ~( MONEMU_WIN | MONEMU_TH | MONEMU_CONSOLE );
				}
			mem_alloc (2);
			rom_enable = ROMEN_ASSEM | ROMEN_BASIC;
			mem_set_rom_enable (rom_enable);
			break;
			}
			case  CFG_MTX512:
			{
			if ( cfg.mon_emu & ( MONEMU_WIN | MONEMU_TH | MONEMU_CONSOLE ) )
				{
				mon_term ();
				cfg.mon_emu &= ~( MONEMU_WIN | MONEMU_TH | MONEMU_CONSOLE );
				}
			mem_alloc (4);
			rom_enable = ROMEN_ASSEM | ROMEN_BASIC;
			mem_set_rom_enable (rom_enable);
			break;
			}
			case  CFG_SDX:
			{
			if ( cfg.mon_emu & ( MONEMU_WIN | MONEMU_TH | MONEMU_CONSOLE ) )
				{
				mon_term ();
				cfg.mon_emu &= ~( MONEMU_WIN | MONEMU_TH | MONEMU_CONSOLE );
				}
			mem_alloc (SDX_MEM_BLOCKS);
			rom_enable = ROMEN_ASSEM | ROMEN_BASIC | ROMEN_SDX2;
			mem_set_rom_enable (rom_enable);
			break;
			}
			case  CFG_CPM_MONO:
			{
			mem_alloc (SDX_MEM_BLOCKS);
			rom_enable = ROMEN_ASSEM | ROMEN_BASIC | ROMEN_SDX2 | ROMEN_CPM;
			mem_set_rom_enable (rom_enable);
			if ( cfg.mon_emu != ( MONEMU_WIN | MONEMU_IGNORE_INIT | MONEMU_WIN_MONO ) )
				{
				if ( cfg.mon_emu & ( MONEMU_WIN | MONEMU_TH | MONEMU_CONSOLE ) ) mon_term ();
				cfg.mon_emu = MONEMU_WIN | MONEMU_IGNORE_INIT | MONEMU_WIN_MONO;
				mon_init (cfg.mon_emu, cfg.mon_width_scale, cfg.mon_height_scale);
				}
			break;
			}
			case  CFG_CPM_COLOUR:
			{
			mem_alloc (SDX_MEM_BLOCKS);
			rom_enable = ROMEN_ASSEM | ROMEN_BASIC | ROMEN_SDX2 | ROMEN_CPM;
			mem_set_rom_enable (rom_enable);
			if ( cfg.mon_emu != ( MONEMU_WIN | MONEMU_IGNORE_INIT ) )
				{
				if ( cfg.mon_emu & ( MONEMU_WIN | MONEMU_TH | MONEMU_CONSOLE ) ) mon_term ();
				cfg.mon_emu = MONEMU_WIN | MONEMU_IGNORE_INIT;
				mon_init (cfg.mon_emu, cfg.mon_width_scale, cfg.mon_height_scale);
				}
			break;
			}
			}
		}
	memu_reset ();
	}

static void row_exit_draw (int iState)
	{
	static char *psExit   =  "  Apply   Cancel   Restart   Exit   ";
	static const char *psHelp[EXIT_COUNT]   =
		{
		"<Space> to select: Change disk and/or tape and continue MEMU",
		"<Space> to select: Continue MEMU with no changes",
		"<Space> to select: Make changes and restart MEMU",
		"<Space> to select: Exit MEMU"
		};
	cfg_print (ROW_EXIT, 0, STY_NORMAL, psExit, 0);
	if ( iCfgCur != iCfgOld )
		{
		cfg_print (ROW_EXIT, 0, STY_DISABLED, psExit, 9);
		}
	if ( iState == STATE_FOCUS )
		{
		if ( ( iCfgCur != iCfgOld ) && ( iCfgExit == EXIT_APPLY ) ) iCfgExit =  EXIT_RESTART;
		cfg_print (ROW_EXIT, 9*iCfgExit, STY_HIGHLIGHT, &psExit[9*iCfgExit], 9);
		cfg_help (psHelp[iCfgExit]);
		}
	}

static BOOLEAN row_exit_key (int wk)
	{
	switch (wk)
		{
		case  WK_Left:
		{
		--iCfgExit;
		if ( iCfgExit < 0 )  iCfgExit =  EXIT_COUNT - 1;
		break;
		}
		case  WK_Right:
		{
		++iCfgExit;
		if ( iCfgExit >= EXIT_COUNT ) iCfgExit =  0;
		break;
		}
		case  WK_Return:
		case  ' ':
		{
		switch ( iCfgExit )
			{
			case  EXIT_RESTART:
			{
			cfg_restart ();
			/* This case deliberately runs into the next one */
			}
			case  EXIT_APPLY:
			{
			cfg_set_tape (psCfgTape);
			cfg_set_drive (0, psCfgDirB, psCfgDriveB);
			cfg_set_drive (1, psCfgDirC, psCfgDriveC);
/*
			cfg.kbd_emu = ( cfg.kbd_emu & KBDEMU_COUNTRY ) | ( bCfgRemap ? KBDEMU_REMAP : 0 );
			kbd_init ( cfg.kbd_emu );
*/
			if ( bCfgRemap )
				{
				cfg.kbd_emu	  |= KBDEMU_REMAP;
				kbd_init (cfg.kbd_emu);
				}
			else
				{
				cfg.kbd_emu	  &= ~KBDEMU_REMAP;
				kbd_init (cfg.kbd_emu);
				kbd_apply_unmap ();
				}
			cfg_save (config_fn);
			break;
			}
			case  EXIT_EXIT:
			{
			config_term ();
			terminate ("User exit");
			break;
			}
			}
		bCfgExit =  TRUE;
		break;
		}
		default:
		{
		return   FALSE;
		}
		}
	return   TRUE;
	}

static BOOLEAN row_focus_true (void)
	{
	return   TRUE;
	}

static BOOLEAN row_focus_drive (void)
	{
	return   ( ( iCfgCur == CFG_SDX ) || ( iCfgCur == CFG_CPM_MONO ) || ( iCfgCur == CFG_CPM_COLOUR ) );
	}

struct  s_cfg_ui
	{
	void (*draw) (int);
	BOOLEAN (*key) (int);
	BOOLEAN (*focus) (void);
	}
	cfg_ui[] =
		{
		{ row_cfg_draw, row_cfg_key, row_focus_true },
		{ row_kbd_draw, row_kbd_key, row_focus_true },
		{ row_tape_draw, row_tape_key, row_focus_true },
		{ row_drive_b_draw, row_drive_b_key, row_focus_drive },
		{ row_drive_c_draw, row_drive_c_key, row_focus_drive },
		{ row_exit_draw, row_exit_key, row_focus_true }
		};

void config (void)
	{
	int  nRow   =  sizeof (cfg_ui) / sizeof (cfg_ui[0]);
	int  iFocus =  0;
	int  wk, i;
	cfg_init ();
	bCfgExit =  FALSE;
	bCfgRedraw  =  TRUE;
	while ( ! bCfgExit )
		{
		if ( bCfgRedraw )
			{
			for ( i = 0; i < nRow; ++i )
				{
				cfg_ui[i].draw (cfg_ui[i].focus () ? STATE_NORMAL : STATE_DISABLED);
				}
			bCfgRedraw  =  FALSE;
			}
		cfg_ui[iFocus].draw (STATE_FOCUS);
		wk =  cfg_key ();
		cfg_ui[iFocus].draw (STATE_NORMAL);
		if ( ! cfg_ui[iFocus].key (wk) )
			{
			if ( wk == WK_Down )
				{
				do
					{
					++iFocus;
					if ( iFocus >= nRow )   iFocus   =  0;
					}
				while ( ! cfg_ui[iFocus].focus () );
				}
			else if ( wk == WK_Up )
				{
				do
					{
					--iFocus;
					if ( iFocus < 0 ) iFocus   =  nRow - 1;
					}
				while ( ! cfg_ui[iFocus].focus () );
				}
			}
		}
	config_term ();
	}

BOOLEAN test_cfg_key (int wk)
	{
	switch (wk)
		{
		case WK_Scroll_Lock:
		case WK_Sys_Req:
		{
		config ();
		return TRUE;
		}
		}
	return FALSE;
	}

BOOLEAN read_config (const char *psFile, int *pargc, char ***pargv)
	{
	FILE *pfil;
	char **argv;
	char *psArg;
	char chQuo;
	int  iarg, narg;
	int  ich, nch;
	BOOLEAN  bFirst;
	BOOLEAN  bQuote;
	pfil  =  fopen (psFile, "r");
	if ( pfil == NULL )
		{
		return   FALSE;
		}
	fseek (pfil, 0, SEEK_END);
	nch  =   ftell (pfil);
	fseek (pfil, 0, SEEK_SET);
	psArg =  malloc (nch+1);
	if ( psArg == NULL )
		{
		fclose (pfil);
		return   FALSE;
		}
	if ( fread (psArg, 1, nch, pfil) != nch )
		{
		free (psArg);
		fclose (pfil);
		return   FALSE;
		}
	fclose (pfil);
	psArg[nch]  =  '\0';
	narg     =  0;
	bFirst   =  TRUE;
	bQuote   =  FALSE;
	for ( ich = 0; ich < nch; ++ich )
		{
		if ( ( ! bQuote ) && ( psArg[ich] <= ' ' ) )
			{
			bFirst   =  TRUE;
			}
		else if ( ( bQuote ) && ( psArg[ich] == chQuo ) )
			{
			bQuote   =  FALSE;
			bFirst   =  TRUE;
			}
		else if ( bFirst )
			{
			++narg;
			bFirst   =  FALSE;
			if ( ( psArg[ich] == '"' ) || ( psArg[ich] == '\'' ) )
				{
				bQuote   =  TRUE;
				chQuo    =  psArg[ich];
				}
			}
		}
	if ( narg == 0 )  return   FALSE;
	argv   =  (char **) malloc ((narg+*pargc)*sizeof (char *));
	if ( argv == NULL )
		{
		free (psArg);
		return   FALSE;
		}
	memcpy (argv, *pargv, (*pargc) * sizeof (char *));
	iarg =   *pargc;
	bFirst   =  TRUE;
	bQuote   =  FALSE;
	for ( ich = 0; ich < nch; ++ich )
		{
		if ( ( ! bQuote ) && ( psArg[ich] <= ' ' ) )
			{
			psArg[ich]  =  '\0';
			bFirst   =  TRUE;
			}
		else if ( ( bQuote ) && ( psArg[ich] == chQuo ) )
			{
			psArg[ich]  =  '\0';
			bQuote   =  FALSE;
			bFirst   =  TRUE;
			}
		else if ( bFirst )
			{
			argv[iarg]   =  &psArg[ich];
			bFirst   =  FALSE;
			if ( ( psArg[ich] == '"' ) || ( psArg[ich] == '\'' ) )
				{
				argv[iarg]   =  &psArg[ich+1];
				bQuote   =  TRUE;
				chQuo    =  psArg[ich];
				}
			++iarg;
			}
		}
	*pargc   += narg;
	*pargv   =  argv;
	return   TRUE;
	}

BOOLEAN cfg_options (int *pargc, char ***pargv, int *pi)
	{
	if ( !strcmp((*pargv)[*pi], "-config-file") )
		{
		if ( ++(*pi) == (*pargc) )
			usage();
		config_fn   =  (*pargv)[*pi];
		read_config (config_fn, pargc, pargv);
		}
	else if ( !strcmp((*pargv)[*pi], "-rom-enable") )
		{
		if ( ++(*pi) == (*pargc) )
			usage();
		sscanf ((*pargv)[*pi], "%i", &rom_enable);
		mem_set_rom_enable (rom_enable);
		}
	else if ( !strcmp((*pargv)[*pi], "-mon-size") )
		{
		if ( ++(*pi) == (*pargc) )
			usage();
		sscanf((*pargv)[*pi], "%i", &cfg.mon_height_scale);
		cfg.mon_width_scale = cfg.mon_height_scale/2;
		}
	else if ( !strcmp((*pargv)[*pi], "-disk-dir") )
		{
		if ( ++(*pi) == (*pargc) )
			usage();
		disk_dir =  (*pargv)[*pi];
		}
	else
		{
		return FALSE;
		}
	return TRUE;
	}

void cfg_usage (void)
	{
	fprintf(stderr, "       -config-file file    read configuration options from file\n");
	fprintf(stderr, "       -rom-enable rom_bits bit flags to enable (1) or disable (0) a rom\n");
	fprintf(stderr, "       -mon-size            sets 80 col size (but does not enablen it)\n");
	fprintf(stderr, "       -disk-dir            directory containg disk images\n");
	}

static char *cfg_make_path (const char *psDir, char chDirSep, const char *psFile)
	{
	char * psPath;
	if ( psDir != NULL )
		{
		int nDir = strlen (psDir);
		int nPath = nDir + strlen (psFile) + 2;
		psPath = emalloc (nPath);
		strcpy (psPath, psDir);
		psPath[nDir] = chDirSep;
		strcpy (&psPath[nDir+1], psFile);
		}
	else
		{
		psPath = estrdup (psFile);
		}
	return psPath;
	}

BOOLEAN cfg_test_file (const char *psPath)
	{
	struct stat st;
	int iErr = stat (psPath, &st);
	return ( ( iErr == 0 ) && ( S_ISREG (st.st_mode) ) );
	}

char * cfg_find_file (const char *argv0)
	{
	char *psCfg, *ps1, *ps2;
	char *psPath = getenv ("PATH");
	int nProg = strlen (argv0);
	char *psProg = (char *) emalloc (nProg+4);
	char chPathSep, chDirSep;
	// printf ("argv0 = %s\n", argv0);
	// printf ("psPath = %s\n", psPath);
	// printf ("nProg = %d\n", nProg);

	/* Name of configuration file */

	if ( argv0 == NULL ) return NULL;
	strcpy (psProg, argv0);
	if ( strcasecmp (&psProg[nProg - 4], ".exe") == 0 )
		{
		nProg -= 4;
		psProg[nProg] = '\0';
		}
	strcpy (&psProg[nProg], ".cfg");
	nProg += 4;
	// printf ("psProg = %s\n", psProg);

	/* Guess OS */

	if ( ( strchr (argv0, '\\') != NULL ) || ( ( psPath != NULL ) && ( strchr (psPath, ';' ) != NULL ) ) )
		{
		printf ("Windows OS\n");
		chPathSep = ';';
		chDirSep = '\\';
		}
	else
		{
		printf ("Linux OS\n");
		chPathSep = ':';
		chDirSep = '/';
		}

	/* Direct location */

	if ( strchr (psProg, chDirSep) != NULL )
		{
		if ( cfg_test_file (psProg) )
			{
			// printf ("Found config file.\n");
			return psProg;
			}
		if ( ( psProg[0] == chDirSep ) || ( ( psProg[1] == ':' ) && ( psProg[2] == '\\' ) ) )
			{
			// printf ("Absolute path and no config file.\n");
			free (psProg);
			return NULL;
			}
		}

	/* Search path */

	if ( psPath == NULL ) return NULL;
	psPath = estrdup (psPath);
	ps2 = psPath - 1;
	while ( ps2 != NULL )
		{
		ps1 = ps2 + 1;
		ps2 = strchr (ps1, chPathSep);
		if ( ps2 != NULL ) *ps2 = '\0';
		// printf ("Try folder = %s\n", ps1);
		psCfg = cfg_make_path (ps1, chDirSep, psProg);
		if ( cfg_test_file (psCfg) )
			{
			// printf ("Found %s\n", psCfg);
			free (psPath);
			free (psProg);
			return psCfg;
			}
		free (psCfg);
		}
	free (psPath);
	free (psProg);
	// printf ("No configuration file.\n");
	return NULL;
	}
