/*

ctc.c - Z80 CTC

*/

/*...sincludes:0:*/
#include "types.h"
#include "diag.h"
#include "common.h"
#include "ctc.h"

/*...vtypes\46\h:0:*/
/*...vdiag\46\h:0:*/
/*...vcommon\46\h:0:*/
/*...vctc\46\h:0:*/
/*...e*/

/*...svars:0:*/
#define	CC_INTERRUPT     0x80	/* Interrupts enabled */
#define	CC_COUNTER_MODE  0x40	/* Counter (not timer) mode */
#define	CC_PRESCALER_256 0x20	/* Timer prescale factor is 256, not 16 */
#define	CC_RISING_EDGE   0x10	/* Rising (not falling edge) */
#define	CC_TIMER_TRIGGER 0x08	/* Timer trigger */
#define	CC_CONSTANT      0x04	/* Load of constant is pending */
#define	CC_RESET         0x02	/* Channel has been reset */
#define	CC_CONTROL       0x01	/* Write control register (not int vector) */

typedef struct
	{
	byte control;
	byte prescaler;
	byte constant;
	byte counter;
	BOOLEAN pending;
	} CHANNEL;

#define	N_CHANNELS 4

static CHANNEL ctc_channels[N_CHANNELS];
static byte ctc_int_vector; /* bits 7-3 inclusive */
/*...e*/

/*...sctc_init:0:*/
void ctc_init(void)
	{
	int i;
	ctc_int_vector = 0x00;
	for ( i = 0; i < N_CHANNELS; i++ )
		{
		CHANNEL *c = &(ctc_channels[i]);
		c->control   = (CC_RESET|CC_CONTROL);
		c->prescaler = 0;
		c->constant  = 0; /* treated as 0x100 */
		c->counter   = 0;
		c->pending   = FALSE;
		}
	}
/*...e*/
/*...sctc_in:0:*/
byte ctc_in(int channel)
	{
	byte value = ctc_channels[channel].counter;
	diag_message(DIAG_CTC_REGISTERS, "CTC counter register for channel %d returned as 0x%02x", channel, value);
	return value;
	}
/*...e*/
/*...sctc_out:0:*/
void ctc_out(int channel, byte value)
	{
	CHANNEL *c = &(ctc_channels[channel]);
	if ( c->control & CC_CONSTANT )
		{
		c->control &= ~(CC_CONSTANT|CC_RESET);
		c->constant = value;
			/* 0 is treated as 0x100,
			   The counter is NOT reloaded.
			   Neither is the prescaler. */
		diag_message(DIAG_CTC_REGISTERS, "CTC constant for channel %d set to 0x%02x", channel, value);
		}
	else if ( value & CC_CONTROL )
		{
		c->control = value;
			/* The Zilog Z80 CTC manual says when CC_RESET,
			   the bits in the control register are not changed,
			   but we have changed them. Oops! */
		diag_message(DIAG_CTC_REGISTERS, "CTC control register for channel %d set to 0x%02x", channel, value);
		}
	else
		{
		if ( channel == 0 )
			{
			ctc_int_vector = (value & 0xf8);
			diag_message(DIAG_CTC_REGISTERS, "CTC interrupt vector set to 0x%02x", value);
			}
		else
			diag_message(DIAG_CTC_REGISTERS, "CTC unexpected write of 0x%02x to channel %d !!!", value, channel);
		}
	}
/*...e*/
/*...sctc_reload:0:*/
void ctc_reload(int channel)
	{
	CHANNEL *c = &(ctc_channels[channel]);
	c->counter = c->constant;
	c->prescaler = ( c->control & CC_PRESCALER_256 ) ? 0 : 16; /* Reload prescaler */
	}
/*...e*/
/*...sctc_trigger:0:*/
void ctc_trigger(int channel)
	{
	CHANNEL *c = &(ctc_channels[channel]);
	if ( c->control & CC_COUNTER_MODE )
		if ( (c->control & (CC_CONSTANT|CC_RESET)) == 0 )
			if ( --(c->counter) == 0 )
				{
				c->counter = c->constant; /* Reload */
				if ( c->control & CC_INTERRUPT )
					{
					c->pending = TRUE;
					diag_message(DIAG_CTC_PENDING, "CTC interrupt pending on channel %d", channel);
					}
				}
	}
/*...e*/
/*...sctc_advance:0:*/
/* Important: This advances in terms of the system clock, as fed to
   the whole of the CTC, not in terms of the input to the channel. */

void ctc_advance(int channel, int clks)
	{
	CHANNEL *c = &(ctc_channels[channel]);
	if ( (c->control & CC_COUNTER_MODE) == 0 )
		if ( (c->control & (CC_CONSTANT|CC_RESET)) == 0 )
			while ( clks > 0 )
				{
				int prescaler = c->prescaler == 0 ? 0x100 : c->prescaler;
				if ( clks < prescaler )
					{
					c->prescaler -= (byte) clks;
					break;
					}
				else
					{
					clks -= prescaler;
					c->prescaler = ( c->control & CC_PRESCALER_256 ) ? 0 : 16; /* Reload prescaler */
					if ( --(c->counter) == 0 )
						{
						c->counter = c->constant; /* Reload */
						if ( c->control & CC_INTERRUPT )
							{
							c->pending = TRUE;
							diag_message(DIAG_CTC_PENDING, "CTC interrupt pending on channel %d", channel);
							}
						}
					}
				}
	}
/*...e*/
/*...sctc_int_pending:0:*/
int ctc_int_pending(void)
	{
	int i;
	for ( i = 0; i < N_CHANNELS; i++ )
		{
		CHANNEL *c = &(ctc_channels[i]);
		if ( c->pending )
			{
			c->pending = FALSE;
			diag_message(DIAG_CTC_INTERRUPT, "CTC interrupt noticed on channel %d", i);
			return ctc_int_vector|(i<<1);
			}
		}
	return -1;
	}
/*...e*/
/*...sctc_get_int_vector:0:*/
/* This is provided to enable to PANEL hack to work. */
byte ctc_get_int_vector(void)
	{
	return ctc_int_vector;
	}
/*...e*/
